//*******************************************************
// Conversion from VRML++ to VRML 2.0 
// Copyright (C) 1996 Stephan Diehl
// 
//
// Converts VRML++(-Parse-Tree) into VRML 2.0(-Parse-Tree)
//
//*******************************************************

#include <iostream.h>
#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include <list.h>
#include <stack.h>
#include <string.h>

#include "vrmlnode.h"

#include "tokens.h"


char* selfname;

#define replace_kid(kids,i,new) { kids.replace(i,i,new); }

ParseTreeNode* process_declaration(ParseTreeNode*, list<VrmlNodeType*>,
                                   list<VrmlNodeType::NameTypeRec*>*);
ParseTreeNode* process_definition(ParseTreeNode*, list<VrmlNodeType*>,
                                  list<VrmlNodeType::NameTypeRec*>);
void add_dec(int, list<VrmlNodeType::NameTypeRec*>, ParseTreeNode*,
             list<VrmlNodeType::NameTypeRec*>*);

char* nodeTypeToString(int);
char* nodeTypeInitValue(int);

/* The constants defined in parsetree.h (0 to ...) and
   tokens.h (258 to ...) have to be disjoined !!! */

int hasSupers=0;
list<ParseTreeNode*> dynRoutes;
list<ParseTreeNode*> dynNodeNames;
list<ParseTreeNode*> dynEvents;


int is_pt_route_event_b(int t)
{ switch (t)
   { case PT_ROUTE_EVENT_F_SF:
     case PT_ROUTE_EVENT_F_MF:
     case PT_ROUTE_EVENT_EX_SF:
     case PT_ROUTE_EVENT_EX_MF:
     case PT_ROUTE_EVENT_EI_SF:
     case PT_ROUTE_EVENT_EI_MF:
     case PT_ROUTE_EVENT_EO_SF:
     case PT_ROUTE_EVENT_EO_MF:
     case PT_R_EVENT_B:	return 1;
    }
  return 0;
}


ParseTreeNode* convert(ParseTreeNode* node)
 {
  list<ParseTreeNode*> kids=node->getChildren();
  list<ParseTreeNode*>::iterator i=kids.begin();
  int is_proto=0;
  switch(node->type)
   { case PT_PROTO: is_proto=1;
     case PT_CLASS: { ParseTreeNode *decl, *parents, *def;
                      VrmlNodeType *t = VrmlNodeType::find(node->name);
                      list<VrmlNodeType*> parentList=(*t).getSuperClasses();
                      list<ParseTreeNode*> ptlist;
                      list<VrmlNodeType::NameTypeRec*> inherited;
                      hasSupers=0;
                      selfname="SELF";
                      decl=*i;
                      if (!is_proto) 
                        { i++;
                          parents=*i;
                          node->type=PT_PROTO;
                        }
                      i++;
                      def=*i;
                      if (!is_proto) 
                         { decl=process_declaration(decl,parentList,&inherited);
                         }
                      def=process_definition(def,parentList,inherited);
                      ptlist.push_back(decl);
                      ptlist.push_back(def);
                      node->setChildren(ptlist);
                      break;
                    }
     case PT_USE: if (strcmp(node->name,"SELF")==0) node->name=selfname;
                  break;
     case PT_UNROUTE:
     case PT_ROUTE: { ParseTreeNode *from, *to;
                      from=(*i);
                      i++;
                      to=(*i);
                       
                      if (from->type==PT_R_EVENT_C)
                        if (strcmp(from->name,"SELF")==0) from->name=selfname;
                      if (to->type==PT_R_EVENT_C)
                        if (strcmp(to->name,"SELF")==0) to->name=selfname;

                      if ( (from->type!=PT_R_EVENT_C)
                          || (to->type!=PT_R_EVENT_C) )
                        { dynRoutes.push_back(node);
                          if (from->type==PT_R_EVENT_A)
                             { dynEvents.push_back(from);
                             }
                          if (to->type==PT_R_EVENT_A)
                             { dynEvents.push_back(to);
                             }
                          if (is_pt_route_event_b(from->type))
                           { dynEvents.push_back(from);
                           }
                         if (is_pt_route_event_b(to->type))
                           { dynEvents.push_back(to);
                           }
                         if (from->type==PT_R_EVENT_C)
                           { dynNodeNames.push_back(from);
                           }
                         if (to->type==PT_R_EVENT_C)
                           { dynNodeNames.push_back(to);
                           }
                         node=pt_node(PT_NONE,NULL);
                        }
                      break;
                    }
                
     default: 	if (!kids.empty())
                 { list<ParseTreeNode*> ptlist;
                   for(;i!=kids.end();i++)
                   { ptlist.push_back(convert(*i));
                   }
                   node->setChildren(ptlist);
                 }
   }
 
     return node;
 }

ParseTreeNode* process_declaration(ParseTreeNode* dec, 
                                   list<VrmlNodeType*> parentList,
                                   list<VrmlNodeType::NameTypeRec*>* inherited)
 { list<VrmlNodeType*>::iterator pi=parentList.begin(); 
   // add all fields and events of the parents, which have not been
   // present in the declaration
   for(;pi!=parentList.end();pi++)
    { if (strcmp((*pi)->getName(),"SFNode")!=0)
       { VrmlNodeType *t = VrmlNodeType::find((*pi)->getName());
         hasSupers=1;
         if (t!=NULL) 
          { add_dec(PT_FIELD, (*t).getFields(),dec,inherited);
            add_dec(PT_EVENTIN, (*t).getEventIns(),dec,inherited);
            add_dec(PT_EVENTOUT, (*t).getEventOuts(),dec,inherited);
          }
         else
          { cerr << "Error: superclass " << (*pi)->getName() << " not found !\n";
          }
         if ((*t).isClass())
            { process_declaration(dec,(*t).getSuperClasses(),inherited);
            }
       }
    }
   return dec;
 }

ParseTreeNode* process_definition(ParseTreeNode* def, 
                                  list<VrmlNodeType*> parentList,
                                  list<VrmlNodeType::NameTypeRec*> inherited)
 { list<ParseTreeNode*> kids=def->getChildren();
   list<ParseTreeNode*>::iterator i=kids.begin();
   list<ParseTreeNode*> newkids;
   ParseTreeNode *result;
 
   int selfDefined=0;
   
   dynRoutes.erase(dynRoutes.begin(),dynRoutes.end());
   dynNodeNames.erase(dynNodeNames.begin(),dynNodeNames.end());
   dynEvents.erase(dynEvents.begin(),dynEvents.end());

   if (hasSupers)
     { // add nodes of each superclass and use IS to propagate fields 
       // and events

       list<VrmlNodeType::NameTypeRec*>::iterator j=inherited.begin();
       list<VrmlNodeType*>::iterator p=parentList.begin();
       ParseTreeNode *node,*defs;
       
       for(;p!=parentList.end();p++)
        { list<ParseTreeNode*> inhGuts;
          int found=1;
          defs=pt_node(PT_GUTS,NULL);
         // we have two choices: we propagate each event or field 
         //  * only to the first class which supports it
         //  * to all classes which support it
         // we chose the first option
         for(;((found) && (j!=inherited.end()));j++)
            { char* name=(*j)->name;
              if (   ((*p)->hasEventIn(name))
                  || ((*p)->hasEventOut(name))
                  || ((*p)->hasField(name))
                  || ((*p)->hasExposedField(name)))
                {
                  inhGuts.push_back(pt_node(PT_IDENTIFIER,
                                             name,
                                             pt_node(PT_IS,
                                                     NULL,
                                                     pt_node(PT_R_EVENT_A,name))));
                }
               else { found=0; j--; } 
             }
             defs->setChildren(inhGuts);
             node=pt_node(PT_NODE,(*p)->getName(),defs);
             if (!selfDefined)
                 { result=pt_node(PT_DEF,"SELF",node);
                   selfDefined=1;
                 }
                else
                 result=node;
            newkids.push_back(result);
        }
       
     }
   for(;i!=kids.end();i++)
      { ParseTreeNode* result;
        if (!selfDefined)
           switch((*i)->type)
            { case PT_DEF: { //ParseTreeNode *node, *pt, *ptuse;
                             //list<ParseTreeNode*> ptlist;
                             //node=((*i)->getChildren()).front();
                             //pt=pt_node(PT_DEF,"SELF",node);
                             //ptuse=pt_node(PT_USE,"SELF");
                             //ptlist.push_back(ptuse);
                             //(*i)->setChildren(ptlist);
                             //newkids.push_back(pt);
                             selfname=(*i)->name;
                             result=(*i);
                             break;
                           }
              case PT_USE:  //result=pt_node(PT_DEF,"SELF",*i);
                            selfname=(*i)->name;
                            result=(*i);
                            break;
              case PT_NODE: result=pt_node(PT_DEF,"SELF",*i);
                            break;
              default: result=convert(*i);
             }
          else
            result=convert(*i);
         
        newkids.push_back(result);
      }

   if (!dynRoutes.empty())
      { ParseTreeNode *dynNode, *routes, *names, *events;
        
        routes=pt_node(PT_DROUTES,NULL);
        routes->setChildren(dynRoutes);
        names=pt_node(PT_DNAMES,NULL);
        names->setChildren(dynNodeNames);
        events=pt_node(PT_DEVENTS,NULL);
        events->setChildren(dynEvents);
        dynNode=pt_node(PT_DYN_ROUTE,NULL,routes,names,events);
        newkids.push_back(dynNode);
      }  
   def->setChildren(newkids);
   return def;
 }

char* nodeTypeInitValue(int type)
 { switch (type)
   { case SFBOOL: return "TRUE";
     case SFCOLOR:
     case MFCOLOR: return "0 0 0";
     case SFFLOAT:
     case MFFLOAT: return "0";
     case SFIMAGE: return "0 0 0";
     case SFINT32: 
     case MFINT32: return "0";
     case SFROTATION:
     case MFROTATION: return "0 0 0 0";
     case SFSTRING:
     case MFSTRING: return "\"\"";
     case SFTIME: return "0";
     case SFVEC2F:
     case MFVEC2F:return "0 0";
     case SFVEC3F:
     case MFVEC3F: return "0 0 0";
     default:	return("NULL");
            	break;
    }
   return("NULL");
  }

void add_dec(int type, list<VrmlNodeType::NameTypeRec*> fields, 
             ParseTreeNode* dec, 
             list<VrmlNodeType::NameTypeRec*>* inherited)
 { list<VrmlNodeType::NameTypeRec*>::iterator fi=fields.begin();
   list<ParseTreeNode*> decList=dec->getChildren();
   list<ParseTreeNode*>::iterator di=decList.begin(); 
   for(;fi!=fields.end();fi++)
    { int found=0;
      for(di=decList.begin();((!found) && (di!=decList.end()));di++)
       { if (strcmp((*di)->name,(*fi)->name)==0) 
           {found=1;}
       }
      if (!found)
       { ParseTreeNode* fieldType=pt_node(PT_ID,
                                          nodeTypeToString((*fi)->type));
         ParseTreeNode* newDec;
         if ((type==PT_FIELD) || (type==PT_EXPFIELD))
           { ParseTreeNode* initVal;
             if ((*fi)->pt) initVal=(*fi)->pt;
                else initVal=pt_node(PT_ID,nodeTypeInitValue((*fi)->type));
             newDec=pt_node(type,(*fi)->name,fieldType,initVal);
           }
            else
           { newDec=pt_node(type,(*fi)->name,fieldType);
           }
         decList.push_back(newDec);
         inherited->push_back(*fi);
       }
    }
   dec->setChildren(decList);
   
 }

char* nodeTypeToString(int type)
{ switch (type)
  { case SFBOOL: return("SFBool");
    case SFCOLOR: return("SFColor");
    case SFFLOAT: return("SFFloat");
    case SFIMAGE: return("SFImage");
    case SFINT32: return("SFInt32");
    case SFNODE: return("SFNode");
    case SFROTATION: return("SFRotation");
    case SFSTRING: return("SFString");
    case SFTIME: return("SFTime");
    case SFVEC2F: return("SFVec2f");
    case SFVEC3F: return("SFVec3f");
    case MFCOLOR: return("MFColor");
    case MFFLOAT: return("MFFloat");
    case MFINT32: return("MFInt32");
    case MFNODE: return("MFNode");
    case MFROTATION: return("MFRotation");
    case MFSTRING: return("MFString");
    case MFVEC2F: return("MFVec2f");
    case MFVEC3F: return("MFVec3f");
    default: return("SFNode");
   }
 return("SFNode");
}