/**************************************************
 * VRML++ Preprocessor
 * Copyright (C) 1996 Stephan Diehl
 **************************************************
 */


//
// The ParseTreeNode class is responsible for storing information about 
// each node in the parse-tree
//

#include <list.h>

class ParseTreeNode {
  public:
    // Constructor.  Takes name of type 
    // Copies the string given as name.
    ParseTreeNode(const char *nm);
    ParseTreeNode(const char *nm, int type);

    // Destructor exists mainly to deallocate storage for name
    ~ParseTreeNode();
    
    // Routines for adding/getting children
    void addChild(ParseTreeNode *);
    void addChild_front(ParseTreeNode *);
    void setChildren(list<ParseTreeNode*> l);

    list<ParseTreeNode*> getChildren(void);
 
    const char *getName() const { return name; }

    char *name;

    int type;
    
    int ftype;

    char *classname;

  private:
    // Children are stored in a list
    list<ParseTreeNode*> children;
};

/* Here we define the types used in the parsetree */

#define PT_ID            0
#define PT_DECS          1
#define PT_DEF           2
#define PT_USE           3
#define PT_PROTO         4
#define PT_EXTPROTO      5
#define PT_INTERFACE     6
#define PT_EVENTIN       7
#define PT_EVENTOUT      8
#define PT_FIELD         9
#define PT_EXPFIELD      10
#define PT_EXTINTERFACE  11
#define PT_ROUTE         12
#define PT_R_EVENT_A     13
#define PT_R_EVENT_B     14
#define PT_NODE          15
#define PT_GUTS          16
#define PT_IDENTIFIER    17
#define PT_EVENTIN_IS    18
#define PT_EVENTOUT_IS   19
#define PT_CLASS         20
#define PT_EXTCLASS      21
#define PT_SUPERS        22
#define PT_VALUE         23
#define PT_SFNODE        24
#define PT_SFN_NULL      25
#define PT_MFNODE        26
#define PT_MFNODE2       27
#define PT_IS            28
#define PT_R_EVENT_C     29
#define PT_NONE          30
#define PT_DYN_ROUTE     31
#define PT_DROUTES       32
#define PT_DNAMES        33
#define PT_DEVENTS       34
#define PT_ROUTE_EVENT_F_SF 35
#define PT_ROUTE_EVENT_F_MF 36
#define PT_ROUTE_EVENT_EX_SF 37
#define PT_ROUTE_EVENT_EX_MF 38
#define PT_ROUTE_EVENT_EI_SF 39
#define PT_ROUTE_EVENT_EI_MF 40
#define PT_ROUTE_EVENT_EO_SF 41
#define PT_ROUTE_EVENT_EO_MF 42
#define PT_UNROUTE 43

ParseTreeNode* pt_node(int, const char*);
ParseTreeNode* pt_node(int, const char*, ParseTreeNode*);
ParseTreeNode* pt_node(int, const char*, ParseTreeNode*, ParseTreeNode*);
ParseTreeNode* pt_node(int, const char*, ParseTreeNode*, ParseTreeNode*, 
                                   ParseTreeNode*);
ParseTreeNode* pt_node(int, const char*, ParseTreeNode*, ParseTreeNode*, 
                                   ParseTreeNode*, ParseTreeNode*);

void unparse(ParseTreeNode*);

ParseTreeNode* convert(ParseTreeNode*);


